package de.fhdw.wtf.common.ast;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.AttributeModifierVisitorReturn;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.token.Token;

/**
 * 
 * This class represents an attribute of a class. It is held by a ClassType, can contain multiple
 * {@link AttributeModifier} and must contain a {@link Type}, which determines whether this attribute is a list or a
 * single value.
 */
public final class Attribute extends SyntaxObject {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = 1438824751069001192L;
	
	/**
	 * The name of the attribute.
	 */
	private final String name;
	
	/**
	 * The type of the attribute.
	 */
	private final Type attrType;
	
	/**
	 * The modifiers of the attribute.
	 */
	private final Collection<AttributeModifier> modifiers;
	
	/**
	 * Holds the database type id, if available.
	 */
	private DatabaseIDState attributeId;
	
	/**
	 * Private Constructor of {@link Attribute}.
	 * 
	 * @param name
	 *            The name of the attribute.
	 * @param attrType
	 *            The type of the attribute.
	 * @param modifiers
	 *            The modifiers of the attribute.
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	private Attribute(final String name,
			final Type attrType,
			final Collection<AttributeModifier> modifiers,
			final Token firstToken,
			final Token lastToken) {
		super(firstToken, lastToken);
		this.name = name;
		this.modifiers = modifiers;
		this.attrType = attrType;
		this.setAttributeId(new DatabaseIDNotSetState());
	}
	
	/**
	 * Creates a {@link Attribute}-Object.
	 * 
	 * @param name
	 *            The name of the attribute.
	 * @param attrType
	 *            The type of the attribute.
	 * @param modifiers
	 *            The modifiers of the attribute.
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return The Attribute-Object.
	 */
	public static Attribute create(final String name,
			final Type attrType,
			final Collection<AttributeModifier> modifiers,
			final Token firstToken,
			final Token lastToken) {
		return new Attribute(name, attrType, modifiers, firstToken, lastToken);
	}
	
	/**
	 * Creates a {@link Attribute}-Object.
	 * 
	 * @param name
	 *            The name of the attribute.
	 * @param attrType
	 *            The type of the attribute.
	 * @param modifiers
	 *            The modifiers of the attribute.
	 * @param firstToken
	 *            firstToken
	 * @return The Attribute-Object.
	 */
	public static Attribute create(final String name,
			final Type attrType,
			final Collection<AttributeModifier> modifiers,
			final Token firstToken) {
		return new Attribute(name, attrType, modifiers, firstToken, null);
	}
	
	/**
	 * Returns true if <code>this</code> is a findable attribute. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isFindable() {
		final AttributeModifierVisitorReturn<Boolean> visitor = new AttributeModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final AttributeModifierFindable modFindable) {
				return true;
			}
			
			@Override
			public Boolean handle(final AttributeModifierPrior modPrior) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierMutable modMutable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierSymmetric modSymmetric) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a prior attribute. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isPrior() {
		final AttributeModifierVisitorReturn<Boolean> visitor = new AttributeModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final AttributeModifierFindable modFindable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierPrior modPrior) {
				return true;
			}
			
			@Override
			public Boolean handle(final AttributeModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierMutable modMutable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierSymmetric modSymmetric) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a transient attribute. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isTransient() {
		final AttributeModifierVisitorReturn<Boolean> visitor = new AttributeModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final AttributeModifierFindable modFindable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierPrior modPrior) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierTransient modTransient) {
				return true;
			}
			
			@Override
			public Boolean handle(final AttributeModifierMutable modMutable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierSymmetric modSymmetric) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a mutable attribute. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isMutable() {
		final AttributeModifierVisitorReturn<Boolean> visitor = new AttributeModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final AttributeModifierFindable modFindable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierPrior modPrior) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierMutable modMutable) {
				return true;
			}
			
			@Override
			public Boolean handle(final AttributeModifierSymmetric modSymmetric) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a symmetric attribute. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isSymmetric() {
		final AttributeModifierVisitorReturn<Boolean> visitor = new AttributeModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final AttributeModifierFindable modFindable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierPrior modPrior) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierMutable modMutable) {
				return false;
			}
			
			@Override
			public Boolean handle(final AttributeModifierSymmetric modSymmetric) {
				return true;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Processes a Returnvisitor with Return-Type Boolean and returns the result.
	 * 
	 * @param visitor
	 *            visitor.
	 * @return boolean.
	 */
	private boolean processModifierVisitor(final AttributeModifierVisitorReturn<Boolean> visitor) {
		for (final AttributeModifier current : this.getModifiers()) {
			if (current.accept(visitor)) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * @return the name
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * @return the AttributeType
	 */
	public Type getAttrType() {
		return this.attrType;
	}
	
	/**
	 * Returns the local collection of {@link AttributeModifier}. The list can't be modified.
	 * 
	 * @return the modifiers
	 */
	public Collection<AttributeModifier> getModifiers() {
		return Collections.unmodifiableCollection(this.modifiers);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Attribute) {
			boolean result = true;
			final Attribute other = (Attribute) o;
			
			result =
					result && this.getName().equals(other.getName()) && this.getAttrType().equals(other.getAttrType())
							&& this.getModifiers().size() == other.getModifiers().size();
			if (result) {
				final AttributeModifier[] thisModifier =
						this.getModifiers().toArray(new AttributeModifier[this.getModifiers().size()]);
				final AttributeModifier[] otherModifier =
						other.getModifiers().toArray(new AttributeModifier[other.getModifiers().size()]);
				for (int i = 0; i < this.getModifiers().size(); i++) {
					result = result && thisModifier[i].equals(otherModifier[i]);
				}
				return result;
			}
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = this.getName().hashCode() ^ this.getAttrType().hashCode();
		for (final AttributeModifier modifier : this.getModifiers()) {
			result ^= modifier.hashCode();
		}
		return result;
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getName()).append(AstDescriptionConstants.COLON_TOKEN).append(this.getAttrType().toString());
		if (!this.getModifiers().isEmpty()) {
			final Iterator<AttributeModifier> i = this.getModifiers().iterator();
			while (i.hasNext()) {
				final AttributeModifier current = i.next();
				result.append(' ').append(current.toString());
			}
		}
		result.append(AstDescriptionConstants.SEMICOLON_TOKEN);
		return result.toString();
	}
	
	/**
	 * Returns the DatabaseIDState of this.
	 * 
	 * @return DatabaseIDState
	 */
	public DatabaseIDState getAttributeId() {
		return this.attributeId;
	}
	
	/**
	 * Set the DatabaseIDState of this.
	 * 
	 * @param attributeId
	 *            DatabaseIDState
	 */
	public void setAttributeId(final DatabaseIDState attributeId) {
		this.attributeId = attributeId;
	}
	
}
