package de.fhdw.wtf.common.ast.type;

import de.fhdw.wtf.common.ast.DatabaseIDState;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturnException;
import de.fhdw.wtf.common.exception.ast.ASTException;

/**
 * This class represents a state of a {@link TypeProxy}, in which a type is referenced by its name. After the referencer
 * has done his work, the state will change into an {@link InvalidState} or a {@link ByReferenceState}.
 */
public final class ByNameState implements TypeProxyState {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -8948427306855242097L;
	
	/**
	 * The Name of the possible target.
	 */
	private final Name name;
	
	/**
	 * Constructor of {@link ByNameState}.
	 * 
	 * @param name
	 *            name
	 */
	private ByNameState(final Name name) {
		this.name = name;
	}
	
	/**
	 * Creates a {@link ByNameState}-Object.
	 * 
	 * @param name
	 *            name
	 * @return The ByNameState-Object
	 */
	public static ByNameState create(final Name name) {
		return new ByNameState(name);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof ByNameState) {
			final ByNameState other = (ByNameState) o;
			return this.getName().equals(other.getName());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getName().hashCode();
	}
	
	@Override
	public String toString() {
		return this.getName().toString();
		// return ByNameState.class.getSimpleName();
	}
	
	// //// PROJECTIONS //////
	
	/**
	 * Get the Name of the target.
	 * 
	 * @return name
	 */
	public Name getName() {
		return this.name;
	}
	
	@Override
	public String getTypeString() {
		return this.toString();
	}
	
	@Override
	public void accept(final TypeProxyStateVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final TypeProxyStateVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final TypeProxyStateVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final TypeProxyStateVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public int compareTo(final TypeProxyState o) {
		return o.accept(new TypeProxyStateVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final ByNameState byName) {
				return ByNameState.this.getName().toString().compareTo(byName.getName().toString()) * -1;
			}
			
			@Override
			public Integer handle(final InvalidState invalidState) {
				return 1;
			}
			
			@Override
			public Integer handle(final ByReferenceState byType) {
				return -1;
			}
		});
	}
	
	@Override
	public DatabaseIDState getTypeId() {
		throw new ASTException("TypeProxy: cannot get database ID when referencing type by name");
	}
	
	@Override
	public void setTypeId(final DatabaseIDState typeId) {
		throw new ASTException("TypeProxy: cannot set database ID when referencing type by name");
	}
	
}
