package de.fhdw.wtf.common.ast.type;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.ConstructorByTypeAndSignatureState;
import de.fhdw.wtf.common.ast.GroupElement;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.ClassModifierVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.GroupElementExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.GroupElementReturnVisitor;
import de.fhdw.wtf.common.ast.visitor.GroupElementVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.Token;
import de.fhdw.wtf.walker.walker.HelperUtils;

/**
 * This class represents a class for code generation. Such a {@link ClassType} is always held by a <Group>, can contain
 * multiple {@link ClassModifier} , multiple {@link Attribute}s, multiple {@link Operation}s and extend other
 * {@link ClassType}es.
 */
public abstract class ClassType extends AtomicType implements GroupElement {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -2421172719092129548L;
	
	/**
	 * Line-Break.
	 */
	private static final String BREAK = " \n";
	
	/**
	 * The modifiers of this {@link ClassType}.
	 */
	private final Collection<ClassModifier> modifiers;
	
	/**
	 * The attributes of this {@link ClassType}.
	 */
	private final Collection<Attribute> attributes;
	
	/**
	 * The subtypes of this {@link ClassType}.
	 */
	private final Collection<ClassType> subTypes;
	
	/**
	 * The operations of this {@link ClassType}.
	 */
	private final Collection<Operation> operations;
	
	/**
	 * The constructors of this {@link ClassType}.
	 */
	protected final Collection<Constructor> constructors;
	
	/**
	 * Constructor of {@link ClassType}.
	 * 
	 * @param name
	 *            The name of the {@link ClassType}.
	 * @param modifiers
	 *            The modifiers of the {@link ClassType}.
	 * @param attributes
	 *            The attributes of the {@link ClassType}.
	 * @param superTypes
	 *            The supertypes of the {@link ClassType}.
	 * @param operations
	 *            The operations of the {@link ClassType}.
	 * @param constructors
	 *            The constructors called by {@link ClassType}.
	 * @param firstToken
	 *            firstToken
	 * @param subTypes
	 *            The subtypes of the {@link ClassType}.
	 * @param lastToken
	 *            lastToken
	 */
	public ClassType(final Name name,
			final Collection<ClassModifier> modifiers,
			final Collection<Attribute> attributes,
			final Collection<Type> superTypes,
			final Collection<Operation> operations,
			final Collection<Constructor> constructors,
			final Token firstToken,
			final Collection<ClassType> subTypes,
			final Token lastToken) {
		super(name, firstToken, lastToken, superTypes);
		this.modifiers = modifiers;
		this.attributes = attributes;
		this.subTypes = subTypes;
		this.operations = operations;
		this.constructors = constructors;
	}
	
	/**
	 * Returns true if <code>this</code> is an abstract class. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isAbstract() {
		final ClassModifierVisitorReturn<Boolean> visitor = new ClassModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final ClassModifierAbstract modAbstract) {
				return true;
			}
			
			@Override
			public Boolean handle(final ClassModifierService modService) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierVisitable visitable) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a service class. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isService() {
		final ClassModifierVisitorReturn<Boolean> visitor = new ClassModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final ClassModifierAbstract modAbstract) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierService modService) {
				return true;
			}
			
			@Override
			public Boolean handle(final ClassModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierVisitable visitable) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a transient class. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isTransient() {
		final ClassModifierVisitorReturn<Boolean> visitor = new ClassModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final ClassModifierAbstract modAbstract) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierService modService) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierTransient modTransient) {
				return true;
			}
			
			@Override
			public Boolean handle(final ClassModifierVisitable visitable) {
				return false;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Returns true if <code>this</code> is a visitable class. False otherwise.
	 * 
	 * @return boolean.
	 */
	public boolean isVisitable() {
		final ClassModifierVisitorReturn<Boolean> visitor = new ClassModifierVisitorReturn<Boolean>() {
			@Override
			public Boolean handle(final ClassModifierAbstract modAbstract) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierService modService) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierTransient modTransient) {
				return false;
			}
			
			@Override
			public Boolean handle(final ClassModifierVisitable visitable) {
				return true;
			}
		};
		return this.processModifierVisitor(visitor);
	}
	
	/**
	 * Processes a Return Visitor with Return-Type Boolean and returns the result.
	 * 
	 * @param visitor
	 *            visitor.
	 * @return boolean.
	 */
	private boolean processModifierVisitor(final ClassModifierVisitorReturn<Boolean> visitor) {
		for (final ClassModifier current : this.getModifiers()) {
			if (current.accept(visitor)) {
				return true;
			}
		}
		return false;
	}
	
	@Override
	public boolean equals(final Object o) {
		if (!super.equals(o)) {
			return false;
		}
		if (o instanceof TypeProxy) {
			return o.equals(this);
		}
		if (o instanceof ClassType) {
			final ClassType other = (ClassType) o;
			
			boolean result =
					this.getAttributes().size() == other.getAttributes().size()
							&& this.getModifiers().size() == other.getModifiers().size()
							&& this.getOperations().size() == other.getOperations().size()
							&& this.getSubTypes().size() == other.getSubTypes().size()
							&& this.getSuperTypes().size() == other.getSuperTypes().size()
							&& this.getConstructors().size() == other.getConstructors().size();
			if (result) {
				for (final Attribute current : this.getAttributes()) {
					result = result && other.getAttributes().contains(current);
					if (!result) {
						return false;
					}
				}
				for (final ClassModifier current : this.getModifiers()) {
					result = result && other.getModifiers().contains(current);
					if (!result) {
						return false;
					}
				}
				for (final Type current : this.getSuperTypes()) {
					result = result && other.getSuperTypes().contains(current);
					if (!result) {
						return false;
					}
				}
				for (final Operation current : this.getOperations()) {
					result = result && other.getOperations().contains(current);
					if (!result) {
						return false;
					}
				}
				for (final ClassType current : this.getSubTypes()) {
					result = result && other.getSubTypes().contains(current);
					if (!result) {
						return false;
					}
				}
				for (final Constructor current : this.getConstructors()) {
					result = result && other.getConstructors().contains(current);
					if (!result) {
						return false;
					}
				}
				return true;
			}
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = super.hashCode();
		for (final Attribute attribute : this.getAttributes()) {
			result ^= attribute.hashCode();
		}
		for (final ClassModifier modifier : this.getModifiers()) {
			result ^= modifier.hashCode();
		}
		for (final Type current : this.getSuperTypes()) {
			result ^= current.hashCode();
		}
		for (final Operation operation : this.getOperations()) {
			result ^= operation.hashCode();
		}
		for (final ClassType classType : this.getSubTypes()) {
			result ^= classType.hashCode();
		}
		for (final Constructor constructor : this.getConstructors()) {
			result ^= constructor.hashCode();
		}
		return result;
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getName()).append(AstDescriptionConstants.COLON_TOKEN);
		// TODO -> Make toString() runnable for RegularClassType AND ExceptionClassType. Leave it here? put it down to
		// RegClassType and ExcClassType?
		result.append(AstDescriptionConstants.CLASS_TOKEN);
		
		result.append(AstDescriptionConstants.EQUAL_TOKEN);
		if (!this.getSuperTypes().isEmpty()) {
			final Iterator<Type> i = this.getSuperTypes().iterator();
			while (i.hasNext()) {
				final Type current = i.next();
				result.append(current).append(AstDescriptionConstants.PLUS_TOKEN);
			}
		}
		result.append(AstDescriptionConstants.CURLY_BRACKET_OPEN_TOKEN).append(BREAK);
		if (!this.getConstructors().isEmpty()) {
			final Iterator<Constructor> i = this.getConstructors().iterator();
			while (i.hasNext()) {
				final Constructor current = i.next();
				result.append(current).append(BREAK);
			}
		}
		if (!this.getAttributes().isEmpty()) {
			final Iterator<Attribute> i = this.getAttributes().iterator();
			while (i.hasNext()) {
				final Attribute current = i.next();
				result.append(current).append(BREAK);
			}
		}
		if (!this.getOperations().isEmpty()) {
			final Iterator<Operation> i = this.getOperations().iterator();
			while (i.hasNext()) {
				final Operation current = i.next();
				result.append(current).append(BREAK);
			}
		}
		result.append(AstDescriptionConstants.CURLY_BRACKET_CLOSE_TOKEN);
		if (!this.getModifiers().isEmpty()) {
			final Iterator<ClassModifier> i = this.getModifiers().iterator();
			while (i.hasNext()) {
				final ClassModifier current = i.next();
				result.append(' ').append(current);
			}
		}
		result.append(AstDescriptionConstants.SEMICOLON_TOKEN);
		return result.toString();
	}
	
	@Override
	public String getTypeString() {
		return this.getName().toString();
	}
	
	@Override
	public void accept(final GroupElementVisitor visitor) {
		visitor.handle(this);
	}
	
	// ///// PROJECTIONS //////////////
	
	/**
	 * Returns a collection of @link {@link ClassModifier}. The collection can't be modified.
	 * 
	 * @return the modifiers
	 */
	public Collection<ClassModifier> getModifiers() {
		return Collections.unmodifiableCollection(this.modifiers);
	}
	
	/**
	 * Returns a collection of @link {@link Attribute}. The collection can't be modified.
	 * 
	 * @return the attributes
	 */
	public Collection<Attribute> getAttributes() {
		return Collections.unmodifiableCollection(this.attributes);
	}
	
	/**
	 * Returns a collection with all constructors in this {@link ClassType} that have the same parameter-list as
	 * constructorReference. The returned collection will be empty if the type of referencedCOnstructor does not match
	 * this {@link ClassType} or if no matching constructor can be found.
	 * 
	 * @param constructorReferenceState
	 *            the state {@link ConstructorByTypeAndSignatureState} of a constructorReference which might match one
	 *            of the constructors in this class.
	 * @return a collection of all matching constructors.
	 */
	public Collection<Constructor> getMatchingConstructors(final ConstructorByTypeAndSignatureState constructorReferenceState) {
		final Collection<Constructor> result = new LinkedList<>();
		final boolean typeMatches = constructorReferenceState.getType().accept(new TypeVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final AtomicType atomicType) {
				return atomicType.equals(ClassType.this);
			}
			
			@Override
			public Boolean handle(final CompositeType compositeType) {
				return false;
			}
			
			@Override
			public Boolean handle(final TypeProxy typeProxy) {
				return typeProxy.equalsOrTargetEquals(ClassType.this);
			}
		});
		if (typeMatches) {
			final Iterator<Constructor> myConstructors = this.constructors.iterator();
			while (myConstructors.hasNext()) {
				final Constructor currentConstructor = myConstructors.next();
				boolean parameterCheck = true;
				final Iterator<ProductElementType> currentConstructorParameterTypes =
						currentConstructor.getParameters().getElements().iterator();
				final Iterator<Type> referencedConstructorParameterTypes =
						constructorReferenceState.getParamTypes().iterator();
				while (currentConstructorParameterTypes.hasNext()) {
					final ProductElementType currentType = currentConstructorParameterTypes.next();
					if (referencedConstructorParameterTypes.hasNext()) {
						final Type type = referencedConstructorParameterTypes.next();
						if (!currentType.getType().equals(type)) {
							parameterCheck = false;
							break;
						}
					} else {
						parameterCheck = false;
						break;
					}
				}
				if (referencedConstructorParameterTypes.hasNext()) {
					parameterCheck = false;
				}
				
				if (parameterCheck) {
					result.add(currentConstructor);
				}
			}
		}
		return result;
	}
	
	@Override
	public <X extends Exception> void accept(final GroupElementExceptionVisitor<X> visitor) throws X {
		visitor.handle(this);
	}
	
	@Override
	public <X> X accept(final GroupElementReturnVisitor<X> visitor) {
		return visitor.handle(this);
	}
	
	@Override
	public Name getName() {
		return this.getTypeName();
	}
	
	/**
	 * The operations of this {@link ClassType}.
	 * 
	 * @return operations
	 */
	public Collection<Operation> getOperations() {
		return this.operations;
	}
	
	/**
	 * The constructors of this {@link ClassType}.
	 * 
	 * @return constructors
	 */
	public Collection<Constructor> getConstructors() {
		return Collections.unmodifiableCollection(this.constructors);
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return 1;
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return ClassType.this.getTypeName().toString().compareTo(clazz.getTypeName().toString());
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return 1;
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return 1;
			}
		});
	}
	
	@Override
	public void accept(final AtomicTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final AtomicTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	/**
	 * The subtypes of this {@link ClassType}.
	 * 
	 * @return subtypes
	 */
	public Collection<ClassType> getSubTypes() {
		return this.subTypes;
	}
	
	@Override
	public <Y extends Exception> void accept(final AtomicTypeVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final AtomicTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return type.accept(new TypeVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final BaseType baseType) {
						return false;
					}
					
					@Override
					public Boolean handle(final ClassType clazz) {
						if (ClassType.this.getSubTypes().contains(clazz)) {
							return true;
						}
						for (final ClassType subClass : ClassType.this.getSubTypes()) {
							if (subClass.isSupertypeOf(clazz)) {
								return true;
							}
						}
						return false;
					}
				});
			}
			
			@Override
			public Boolean handle(final CompositeType c) {
				return c.accept(new CompositeTypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final ListType list) {
						return false;
					}
					
					@Override
					public Boolean handle(final MapType map) {
						return false;
					}
					
					@Override
					public Boolean handle(final ProductType product) {
						return false;
					}
					
					@Override
					public Boolean handle(final SumType sum) {
						final List<Type> sumElements;
						try {
							final Type standardizedType = sum.standardize();
							sumElements = standardizedType.accept(new TypeVisitorReturn<List<Type>>() {
								
								@Override
								public List<Type> handle(final AtomicType atomicType) {
									return new ArrayList<>();
								}
								
								@Override
								public List<Type> handle(final CompositeType compositeType) {
									return compositeType.accept(new CompositeTypeVisitorReturn<List<Type>>() {
										
										@Override
										public List<Type> handle(final ListType list) {
											return new ArrayList<>();
										}
										
										@Override
										public List<Type> handle(final MapType map) {
											return new ArrayList<>();
										}
										
										@Override
										public List<Type> handle(final ProductType product) {
											return new ArrayList<>();
										}
										
										@Override
										public List<Type> handle(final SumType sum2) {
											return sum2.getElements();
										}
										
										@Override
										public List<Type> handle(final ThrownType thrownType) {
											return new ArrayList<>();
										}
									});
								}
								
								@Override
								public List<Type> handle(final TypeProxy typeProxy) {
									return new ArrayList<>();
								}
							});
						} catch (final SumIsAnythingException e) {
							// If the sum is Anything than this class is not general than anything
							return false;
						}
						
						if (sumElements.size() == 0) {
							return true;
						}
						
						if (sumElements.size() == 1) {
							// If the one SumElement = this class than the sum is special than this class
							if (sumElements.get(0).equals(ClassType.this)) {
								return true;
							}
						}
						
						return ClassType.this.collectionContainsOnlySubTypes(sumElements);
					}
					
					@Override
					public Boolean handle(final ThrownType thrownType) {
						return false;
					}
				});
			}
			
			@Override
			public Boolean handle(final TypeProxy typeProxy) {
				return ClassType.this.isSupertypeOf(HelperUtils.getTargetType(typeProxy));
			}
		});
	}
	
	/**
	 * Returns true, if the collection contains only subtypes of this class. This Method is recursiv so it look in the
	 * subtypes of this class also if in the collection not directly the subtypes of this class.
	 * 
	 * @param collection
	 *            collection of types
	 * @return Boolean
	 */
	boolean collectionContainsOnlySubTypes(final Collection<Type> collection) {
		final Collection<Type> copy = new ArrayList<>();
		copy.addAll(collection);
		
		this.helperMethodForCollectionContainsOnlySubTypes(copy);
		
		return copy.isEmpty();
	}
	
	/**
	 * Helper Method for collectionContainsOnlySubTypes.
	 * 
	 * @param collection
	 *            collection of types
	 */
	private void helperMethodForCollectionContainsOnlySubTypes(final Collection<Type> collection) {
		final Iterator<Type> iterator = collection.iterator();
		while (iterator.hasNext()) {
			final Type current = HelperUtils.getReferencedType(iterator.next());
			for (final ClassType subtype : this.subTypes) {
				if (current.equals(subtype)) {
					iterator.remove();
				}
			}
		}
		
		// look in the subtypes if their subtypes in the collection!
		for (final ClassType subType : this.subTypes) {
			subType.helperMethodForCollectionContainsOnlySubTypes(collection);
		}
	}
	
	@Override
	public ClassType getPrototype() {
		return (ClassType) super.getPrototype();
	}
	
	@Override
	public void setPrototype(final Type prototype) {
		if (!(prototype instanceof ClassType)) {
			throw new ASTException("The prototype \"" + prototype.getClass() + "\" does not match this type \""
					+ this.getClass() + "\"!");
		}
		super.setPrototype(prototype);
	}
	
	/**
	 * Accept-Method from visitor pattern.
	 * 
	 * @param classTypeVisitor
	 *            visitor of type {@link ClassTypeVisitor} to accept
	 */
	public abstract void accept(ClassTypeVisitor classTypeVisitor);
	
	/**
	 * Accept-Method from visitor pattern.
	 * 
	 * @param classTypeVisitorReturn
	 *            visitor of type {@link ClassTypeVisitorReturn} to accept
	 * @return Return type of this return visitor. Type is dependent on generics.
	 */
	public abstract <X> X accept(ClassTypeVisitorReturn<X> classTypeVisitorReturn);
	
	/**
	 * Accept-Method from visitor pattern.
	 * 
	 * @param classTypeVisitorException
	 *            visitor of type {@link ClassTypeVisitorException} to accept
	 * @throws Y
	 *             Exception to throw of this exception visitor. Type is dependent on generics.
	 */
	public abstract <Y extends Exception> void accept(ClassTypeVisitorException<Y> classTypeVisitorException) throws Y;
	
	/**
	 * Accept-Method from visitor pattern.
	 * 
	 * @param classTypeVisitorReturnException
	 *            visitor of type {@link ClassTypeVisitorReturnException} to accept
	 * @return Return type of this return visitor. Type is dependent on generics.
	 * @throws Y
	 *             Exception to throw of this exception visitor. Type is dependent on generics.
	 */
	public abstract <X, Y extends Exception> X accept(ClassTypeVisitorReturnException<X, Y> classTypeVisitorReturnException)
			throws Y;
	
}
