package de.fhdw.wtf.common.ast.type;

import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.Token;

/**
 * A {@link MapType} is an indexed {@link CollectionType}. So it has a key-Type.
 */
public final class MapType extends CollectionType {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -8131265474060721258L;
	
	/**
	 * Type of the key.
	 */
	private final Type key;
	
	/**
	 * Constructor of {@link MapType}.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param key
	 *            Type of the key
	 * @param value
	 *            Type of the value
	 * @param lastToken
	 *            lastToken
	 */
	private MapType(final Token firstToken, final Type key, final Type value, final Token lastToken) {
		super(firstToken, value, lastToken);
		this.key = key;
	}
	
	/**
	 * Creates a {@link MapType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param key
	 *            Type of the key
	 * @param value
	 *            Type of the value
	 * @param lastToken
	 *            lastToken
	 * @return the {@link MapType}-Object
	 */
	public static MapType create(final Token firstToken, final Type key, final Type value, final Token lastToken) {
		return new MapType(firstToken, key, value, lastToken);
	}
	
	/**
	 * Creates a {@link MapType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param key
	 *            Type of the key
	 * @param value
	 *            Type of the value
	 * @return the {@link MapType}-Object
	 */
	public static MapType create(final Token firstToken, final Type key, final Type value) {
		return new MapType(firstToken, key, value, null);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof MapType) {
			final MapType other = (MapType) o;
			return this.getKey().equals(other.getKey()) && this.getOf().equals(other.getOf());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getKey().hashCode() ^ this.getOf().hashCode();
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append('[').append(this.getKey()).append("->").append(this.getOf()).append(']');
		return result.toString();
	}
	
	@Override
	public String getTypeString() {
		final StringBuilder result = new StringBuilder();
		result.append('[').append(this.getKey().getTypeString()).append("->").append(this.getOf().getTypeString())
				.append(']');
		return result.toString();
	}
	
	/**
	 * Type of the key.
	 * 
	 * @return key
	 */
	public Type getKey() {
		return this.key;
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return 1;
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return -1;
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return c.accept(new CompositeTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final ListType list) {
						return 1;
					}
					
					@Override
					public Integer handle(final MapType map) {
						final Type otherKeyReference = map.getKey();
						final Type otherValueReference = map.getOf();
						
						final Integer keyCompare = MapType.this.getKey().compareTo(otherKeyReference);
						final Integer valueCompare = MapType.this.getOf().compareTo(otherValueReference);
						return keyCompare + valueCompare;
					}
					
					@Override
					public Integer handle(final ProductType product) {
						return -1;
					}
					
					@Override
					public Integer handle(final SumType sum) {
						return -1;
					}
					
					@Override
					public Integer handle(final ThrownType thrownType) {
						return 1;
					}
				});
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return 1;
			}
		});
	}
	
	@Override
	public void accept(final CompositeTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final CompositeTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final CompositeTypeVisitorException<X> v) throws X {
		v.handle(this);
		
	}
	
	@Override
	public <X, Y extends Exception> X accept(final CompositeTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return false;
	}
	
	@Override
	public MapType getPrototype() {
		return (MapType) super.getPrototype();
	}
	
	@Override
	public void setPrototype(final Type prototype) {
		if (!(prototype instanceof MapType)) {
			throw new ASTException("The prototype \"" + prototype.getClass() + "\" does not match this type \""
					+ this.getClass() + "\"!");
		}
		super.setPrototype(prototype);
	}
	
	@Override
	public Type standardize() throws SumIsAnythingException {
		return this;
	}
	
	@Override
	public boolean isThrownType() {
		return false;
	}
}
