package de.fhdw.wtf.walker.tasks;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * Analyzes the inheritance trees of the given model. Enriches the model with useful information regarding necessary
 * constructor calls.
 */
public final class AnalyzeInheritanceTreesTask extends SimpleWalkerTask {
	
	/**
	 * Used to identify the type Anything.
	 */
	private static final String ANYTHING = "Anything";
	
	/**
	 * Returns a new instance of {@link AnalyzeInheritanceTreesTask}.
	 * 
	 * @param model
	 *            The model to perform this task on.
	 * @param taskmanager
	 *            The executor which performs this task.
	 * @return a new instance of {@link AnalyzeInheritanceTreesTask}
	 */
	public static AnalyzeInheritanceTreesTask create(final Model model, final TaskExecutor taskmanager) {
		return new AnalyzeInheritanceTreesTask(model, taskmanager);
	}
	
	/**
	 * Instantiates a new instance of {@link AnalyzeInheritanceTreesTask}.
	 * 
	 * @param model
	 *            The model to perform this task on.
	 * @param taskmanager
	 *            The executor which performs this task.
	 */
	private AnalyzeInheritanceTreesTask(final Model model, final TaskExecutor taskmanager) {
		super(model, taskmanager);
	}
	
	/**
	 * Helper class which represents one path along the inheritance-hierarchy of a root class.
	 */
	private class InheritancePath {
		/**
		 * Types ordered along this InheritancePath.
		 */
		private final List<Type> pathElements;
		
		/**
		 * Constructs an empty InheritancePath.
		 */
		private InheritancePath() {
			this.pathElements = new LinkedList<>();
		}
		
		/**
		 * Constructs an InheritancePath with first as its root.
		 * 
		 * @param first
		 *            the first element of this path.
		 */
		InheritancePath(final Type first) {
			this();
			this.addElement(first);
		}
		
		/**
		 * Constructs an InheritancePath with firsts as its path elements.
		 * 
		 * @param firsts
		 *            the path elements for this.
		 */
		InheritancePath(final List<Type> firsts) {
			this();
			this.addElements(firsts);
		}
		
		/**
		 * Constructs an InheritancePath which is identical to firsts.
		 * 
		 * @param firsts
		 *            the path to copy.
		 */
		InheritancePath(final InheritancePath firsts) {
			this(firsts.pathElements);
		}
		
		/**
		 * Adds element to the end of this path.
		 * 
		 * @param element
		 *            the new end of this path.
		 */
		void addElement(final Type element) {
			this.pathElements.add(element);
		}
		
		/**
		 * Adds elements to this path so that these elements remain in the order provided in elements.
		 * 
		 * @param elements
		 *            the elements added to this.
		 */
		void addElements(final List<Type> elements) {
			this.pathElements.addAll(elements);
		}
		
		/**
		 * Returns true if element is also an element of this. False otherwise.
		 * 
		 * @param element
		 *            to check.
		 * @return true if element is part of this. False otherwise.
		 */
		boolean contains(final Type element) {
			return this.pathElements.contains(element);
		}
		
		/**
		 * Returns all elements that this and other have in common. The common elements will be provided the order
		 * provided by this.
		 * 
		 * @param other
		 *            the path to intersect with.
		 * @return common elements in the order of their existence in this path.
		 */
		List<Type> intersect(final InheritancePath other) {
			final List<Type> result = new LinkedList<>();
			final Iterator<Type> myElements = this.pathElements.iterator();
			while (myElements.hasNext()) {
				final Type myElement = myElements.next();
				if (other.contains(myElement)) {
					result.add(myElement);
				}
			}
			return result;
		}
		
		/**
		 * Makes one step in the inheritance-tree. Alternative paths will be returned.
		 * 
		 * @return alternative paths if any exist. An empty llist otherwise.
		 */
		List<InheritancePath> step() {
			final List<InheritancePath> alternativePaths = new LinkedList<>();
			final Type lastElement = this.pathElements.get(this.pathElements.size() - 1);
			final Iterator<Type> followers = lastElement.getSuperTypes().iterator();
			if (followers.hasNext()) {
				final Type firstFollower = followers.next();
				while (followers.hasNext()) {
					final Type follower = followers.next();
					if (!follower.toString().equals(ANYTHING)) {
						final InheritancePath alternativePath = new InheritancePath(this);
						alternativePath.addElement(follower);
						alternativePaths.add(alternativePath);
					}
				}
				if (!firstFollower.toString().equals(ANYTHING)) {
					this.addElement(firstFollower);
				}
			}
			return alternativePaths;
		}
		
		/**
		 * Calculates all inheritance paths reachable from the end of this path and returns them.
		 * 
		 * @return a list with all paths reachable from the last element of this.
		 */
		List<InheritancePath> run() {
			final List<InheritancePath> result = new LinkedList<>();
			final List<InheritancePath> preResult = new LinkedList<>();
			int oldSize = this.pathElements.size();
			List<InheritancePath> alternativePaths = this.step();
			while (this.pathElements.size() != oldSize || !alternativePaths.isEmpty()) {
				preResult.addAll(alternativePaths);
				oldSize = this.pathElements.size();
				alternativePaths = this.step();
			}
			final Iterator<InheritancePath> paths = preResult.iterator();
			while (paths.hasNext()) {
				result.addAll(paths.next().run());
			}
			result.add(this);
			return result;
		}
		
		@Override
		public String toString() {
			return this.pathElements.toString();
		}
		
		@Override
		public boolean equals(final Object o) {
			if (o instanceof InheritancePath) {
				final InheritancePath oAsInheritancePath = (InheritancePath) o;
				return this.pathElements.equals(oAsInheritancePath.pathElements);
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return this.pathElements.hashCode();
		}
		
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		final List<InheritancePath> allPaths = new LinkedList<>();
		final Iterator<Type> superTypes = c.getSuperTypes().iterator();
		while (superTypes.hasNext()) {
			final Type current = superTypes.next();
			if (!current.toString().equals(ANYTHING)) {
				allPaths.addAll(new InheritancePath(current).run());
			}
		}
		this.mapCalls(c, allPaths);
	}
	
	/**
	 * Maps c to all types it has to instantiate during runtime. Uses the map "calls" for this purpose. Maps to an empty
	 * list if c has no responsibilities in the regard of instantiating super types.
	 * 
	 * @param c
	 *            the root of the inheritance tree to map types from.
	 * @param allPaths
	 *            The inheritance tree with c as root in form of all paths that have the direct super types of c as
	 *            root.
	 */
	private void mapCalls(final ClassType c, final List<InheritancePath> allPaths) {
		final Map<Type, Integer> callPriorities = new HashMap<>();
		final List<Type> definitiveResponsibilities = new LinkedList<>();
		final List<InheritancePath> copy = new LinkedList<>();
		copy.addAll(allPaths);
		final Iterator<InheritancePath> copyPaths = copy.iterator();
		while (copyPaths.hasNext()) {
			final InheritancePath copyPath = copyPaths.next();
			final Iterator<InheritancePath> paths = allPaths.iterator();
			while (paths.hasNext()) {
				final InheritancePath path = paths.next();
				final List<Type> interception = path.intersect(copyPath);
				this.gatherResponsibilities(interception, callPriorities, definitiveResponsibilities);
				
			}
			copyPaths.remove();
		}
		this.addToCalls(c, definitiveResponsibilities, callPriorities);
	}
	
	/**
	 * Adds all types in definitiveResponsibilities and direct super types of c as list values to the calls map and uses
	 * c as key. Additionally ensures that the order of the elements in the list will be a valid constructor-call-order
	 * when instantiating c.
	 * 
	 * @param c
	 *            the key to use.
	 * @param definitiveResponsibilities
	 *            those classes that c is definitely responsible to call. Every type in this list must also be a key in
	 *            callPriorities with a valid value != null;
	 * @param callPriorities
	 *            the priorities which are used to determine the call order of the super constructors for c. Will be
	 *            empty after a call to this method!
	 */
	private void addToCalls(final ClassType c,
			final List<Type> definitiveResponsibilities,
			final Map<Type, Integer> callPriorities) {
		final List<Type> callList = new LinkedList<>();
		if (!definitiveResponsibilities.isEmpty()) {
			while (!callPriorities.isEmpty()) {
				final Type max = this.findAndRemoveHighestPathCountType(callPriorities);
				if (definitiveResponsibilities.contains(max)) {
					callList.add(max);
				}
			}
		}
		final Iterator<Type> directSuperTypes = c.getSuperTypes().iterator();
		while (directSuperTypes.hasNext()) {
			final Type superType = directSuperTypes.next();
			if (!superType.toString().equals(ANYTHING) && !definitiveResponsibilities.contains(superType)) {
				callList.add(superType);
			}
		}
		this.getModel().putConstructorCallDependency(c, callList);
	}
	
	/**
	 * Removes the key with the highest integer value as value from callPriorities and returns that key. removes and
	 * returns the first key found if more than one value is at the current maximum. Returns null if callPriorities is
	 * empty.
	 * 
	 * @param callPriorities
	 *            the map to remove a key from.
	 * @return null if callPriorities is empty. The first key with the highest integer as value found in callPriorities.
	 */
	private Type findAndRemoveHighestPathCountType(final Map<Type, Integer> callPriorities) {
		final Iterator<Entry<Type, Integer>> entries = callPriorities.entrySet().iterator();
		int highestPathCount = 0;
		Type highestPathCountType = null;
		while (entries.hasNext()) {
			final Entry<Type, Integer> entry = entries.next();
			final int pathCount = entry.getValue();
			if (pathCount > highestPathCount) {
				highestPathCount = pathCount;
				highestPathCountType = entry.getKey();
			}
		}
		callPriorities.remove(highestPathCountType);
		return highestPathCountType;
	}
	
	/**
	 * Adds all types in intersection to callPriorities and increments the corresponding value for a key by one. Also
	 * adds the first type in intersection to definitiveResponsibilities (no duplicates will be added).
	 * 
	 * @param intersection
	 *            an intersection of two InheritancePaths.
	 * @param callPriorities
	 *            a map with call priorities.
	 * @param definitiveResponsibilities
	 *            a list with types that have to be instantiated by a certain sub type.
	 */
	private void gatherResponsibilities(final List<Type> intersection,
			final Map<Type, Integer> callPriorities,
			final List<Type> definitiveResponsibilities) {
		final Iterator<Type> commonTypes = intersection.iterator();
		if (commonTypes.hasNext()) {
			final Type definitiveOne = commonTypes.next();
			this.addToPriorities(definitiveOne, callPriorities);
			if (!definitiveResponsibilities.contains(definitiveOne)) {
				definitiveResponsibilities.add(definitiveOne);
			}
		}
		while (commonTypes.hasNext()) {
			this.addToPriorities(commonTypes.next(), callPriorities);
		}
	}
	
	/**
	 * Adds the pair (type,2) to callPriorities if type is not a key of callPriorities. Increments the the value of type
	 * by one if type is a key of callPriorities. The initial value is 2 because it represents the amount of paths a
	 * type has been found in.
	 * 
	 * @param type
	 *            the key to add/ to find.
	 * @param callPriorities
	 *            the map too search in.
	 */
	private void addToPriorities(final Type type, final Map<Type, Integer> callPriorities) {
		if (!callPriorities.containsKey(type)) {
			callPriorities.put(type, 1);
		}
		callPriorities.put(type, callPriorities.get(type) + 1);
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void beginTask() throws TaskException {
		// TODO Auto-generated method stub
		
	}
	
}
