package de.fhdw.wtf.walker.tasks;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.AttributeModifier;
import de.fhdw.wtf.common.ast.AttributeModifierSymmetric;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.exception.walker.SymmetricCheckException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.task.TaskExecutorFixed;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * @author hfw413hy
 *
 */
public class SymmetricCheck extends SimpleWalkerTask {
	
	private final HashMap<Group, ArrayList<SymmetricCheckPart>> parts = new HashMap<>();
	private Group currentGroup;
	
	private final void addPart(final Attribute a, final ClassType owner) {
		if (!this.parts.containsKey(this.currentGroup)) {
			this.parts.put(this.currentGroup, new ArrayList<SymmetricCheckPart>());
		}
		this.parts.get(this.currentGroup).add(new SymmetricCheckPart(owner, a));
	}
	
	/**
	 * Erzeugt ein Check-Objekt zum Prüfen, ob symmetrische Beziehungen auf beiden Seiten der Klasse angegeben wurden.
	 * 
	 * @param model
	 *            Das zugrunde liegende Modell, welches geprüft werden soll.
	 * @param taskmanager
	 *            Der Manager.
	 * @return Das Check-Objekt.
	 */
	public static SymmetricCheck create(final Model model, final TaskExecutorFixed taskmanager) {
		return new SymmetricCheck(model, taskmanager);
	}
	
	private SymmetricCheck(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager);
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		// Nichts
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		this.currentGroup = g;
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		if (a.isSymmetric()) {
			this.addPart(a, owner);
		}
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		// Nichts...
	}
	
	/**
	 * 
	 * @param g
	 *            Die Gruppe in der wir symmetrische Beziehungen testen.
	 * @param current
	 *            Das Ausgangs-Attribut.
	 * @throws SymmetricCheckException
	 */
	private void findCorrespondingAttribute(final Group g, final SymmetricCheckPart current)
			throws SymmetricCheckException {
		final Iterator<SymmetricCheckPart> itemIt = this.parts.get(g).iterator();
		while (itemIt.hasNext()) {
			final SymmetricCheckPart cur = itemIt.next();
			
			// Owner-Klassen müssen unterschiedlich sein... sonst zählt dies nicht als "gefunden"
			if (current.getOwner().equals(cur.getOwner())) {
				continue;
			}
			
			// Klasse A: idA : B symmetric(idB);
			// Klasse B: idB : A symmetric(idA);
			
			// Überprüfen der Attributnamen-Referenzierung
			// Prüft ob "idA" (in A) so als Identifier in B (in symmetric(idA)) hinterlegt ist...
			if (!cur.getAttributeName().equals(current.getSymmetricIdentifier())) {
				continue;
			}
			// ...und umgekehrt
			if (!current.getAttributeName().equals(cur.getSymmetricIdentifier())) {
				continue;
			}
			
			// Prüft ob die Typen übereinstimmen
			if (!cur.getAttributeType().equals(current.getOwnerClassName())) {
				continue;
			}
			if (!current.getAttributeType().equals(cur.getOwnerClassName())) {
				continue;
			}
			
			// Wir haben den passenden Typ gefunden der zu "current" passt
			return;
		}
		final String err =
				current.getSymmetricIdentifier() + " : " + current.getOwnerClassName() + " symmetric("
						+ current.getAttributeName() + ");";
		throw SymmetricCheckException.create("Kein passendes Attribut in der Klasse " + current.getAttributeType()
				+ " gefunden! Erwartet: '" + err + "'", current.getAttribute());
	}
	
	/**
	 * Durchläuft und prüft alle Symmetric-Einträge einer Gruppe.
	 * 
	 * @param g
	 *            Die zu prüfende Gruppe
	 * @throws SymmetricCheckException
	 */
	private void finalizeTaskHandleGroup(final Group g) throws SymmetricCheckException {
		final Iterator<SymmetricCheckPart> itemIt = this.parts.get(g).iterator();
		while (itemIt.hasNext()) {
			final SymmetricCheckPart current = itemIt.next();
			this.findCorrespondingAttribute(g, current);
		}
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		final Iterator<Group> groupIt = this.parts.keySet().iterator();
		while (groupIt.hasNext()) {
			this.finalizeTaskHandleGroup(groupIt.next());
		}
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nichts...
	}
	
}

/**
 * 
 * @author hfw413hy
 *
 */
class SymmetricCheckPart {
	private final ClassType Owner;
	private final Attribute OwnerAttribute;
	
	public SymmetricCheckPart(final ClassType owner, final Attribute a) {
		this.Owner = owner;
		this.OwnerAttribute = a;
	}
	
	public ClassType getOwner() {
		return this.Owner;
	}
	
	public Attribute getOwnerAttribute() {
		return this.OwnerAttribute;
	}
	
	public String getSymmetricIdentifier() {
		final Iterator<AttributeModifier> i = this.OwnerAttribute.getModifiers().iterator();
		while (i.hasNext()) {
			final AttributeModifier m = i.next();
			if (m instanceof AttributeModifierSymmetric) {
				final AttributeModifierSymmetric cur = (AttributeModifierSymmetric) m;
				return cur.getIdentifierToken().getIdentifier();
			}
		}
		return "";
	}
	
	public Attribute getAttribute() {
		return this.OwnerAttribute;
	}
	
	public String getAttributeName() {
		return this.OwnerAttribute.getName();
	}
	
	public String getAttributeType() {
		return this.OwnerAttribute.getAttrType().getTypeString();
	}
	
	public String getOwnerClassName() {
		return this.Owner.getName().getLastAddedName().toString();
	}
}
