package de.fhdw.wtf.walker.tasks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.ConstructorByReferenceState;
import de.fhdw.wtf.common.ast.ConstructorByTypeAndSignatureState;
import de.fhdw.wtf.common.ast.ConstructorInvalidState;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.ConstructorReference;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.QualifiedName;
import de.fhdw.wtf.common.ast.Tuple;
import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.ByNameState;
import de.fhdw.wtf.common.ast.type.ByReferenceState;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.InvalidState;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductElementType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnName;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnNameException;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.constants.referencer.BaseTypeConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.referencer.InvalidTypeReferenceException;
import de.fhdw.wtf.common.exception.referencer.ReferencingNotSuccessfulException;
import de.fhdw.wtf.common.exception.referencer.WrongStateException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.token.DummyToken;
import de.fhdw.wtf.common.token.IdentifierToken;

/**
 * This class represents a referencer. A referencer resolves type references by trying to find a {@link Type} for all
 * applied {@link TypeProxy}s, which are referencing by {@link ByNameState}.
 */
public final class TypeReferencer extends Referencer {
	
	/**
	 * Model.
	 */
	private final Model inputModel;
	
	/**
	 * Collection for Exceptions.
	 */
	private final Collection<TaskException> exceptions;
	
	/**
	 * Found Classes.
	 */
	private final List<ClassType> definedClasses;
	
	/**
	 * Found TypeProxys and the classType in which the TypeProxy was found (The classtype need the referencer to know
	 * the hierarchie).
	 */
	private final List<Tuple<TypeProxy, ClassType>> appliedTypeProxys;
	
	/**
	 * Constructor of {@link TypeReferencer}.
	 * 
	 * @param inputModel
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 */
	private TypeReferencer(final Model inputModel, final TaskExecutor taskmanager) {
		super(inputModel, taskmanager);
		this.inputModel = inputModel;
		this.exceptions = new ArrayList<>();
		this.definedClasses = new ArrayList<>();
		this.appliedTypeProxys = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link TypeReferencer}-Object.
	 * 
	 * @param inputModel
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 * @return The Referencer-Object
	 */
	public static TypeReferencer create(final Model inputModel, final TaskExecutor taskmanager) {
		return new TypeReferencer(inputModel, taskmanager);
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do here
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		this.getDefinedClasses().add(c);
		for (final Type superType : c.getSuperTypes()) {
			this.handleType(superType, c);
		}
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		coo.accept(new ConstructorOrOperationExceptionVisitor<TaskException>() {
			@Override
			public void handleOperation(final Operation operation) throws TaskException {
				TypeReferencer.this.handleType(operation.getParameters(), owner);
				TypeReferencer.this.handleType(operation.getReturnType(), owner);
			}
			
			@Override
			public void handleConstructor(final Constructor constructor) throws TaskException {
				TypeReferencer.this.handleType(constructor.getParameters(), owner);
				final Iterator<ConstructorReference> referencedConstructors =
						constructor.getSuperConstructors().iterator();
				while (referencedConstructors.hasNext()) {
					final ConstructorReference current = referencedConstructors.next();
					
					current.getState().accept(new ConstructorReferenceStateVisitor() {
						
						@Override
						public void handle(final ConstructorByReferenceState byReference) {
							TypeReferencer.this.handleType(byReference.getType(), owner);
						}
						
						@Override
						public void handle(final ConstructorInvalidState invalid) {
							// Nothing to do
						}
						
						@Override
						public void handle(final ConstructorByTypeAndSignatureState byName) {
							TypeReferencer.this.handleType(byName.getType(), owner);
							final Iterator<Type> paramsIterator = byName.getParamTypes().iterator();
							while (paramsIterator.hasNext()) {
								TypeReferencer.this.handleType(paramsIterator.next(), owner);
							}
						}
					});
				}
				
			}
		});
		
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		this.handleType(a.getAttrType(), owner);
	}
	
	/**
	 * This operation saves the relevant data from an applied {@link Type}.
	 * 
	 * @param type
	 *            : Type.
	 * @param clazz
	 *            : common.ast.Class.
	 */
	protected void handleType(final Type type, final ClassType clazz) {
		type.accept(new TypeVisitor() {
			@Override
			public void handle(final TypeProxy s) {
				TypeReferencer.this.getAppliedTypeProxys().add(new Tuple<>(s, clazz));
			}
			
			@Override
			public void handle(final CompositeType c) {
				c.accept(new CompositeTypeVisitor() {
					
					@Override
					public void handle(final SumType sum) {
						TypeReferencer.this.getInputModel().getAnything().add(sum);
						final Iterator<Type> i = sum.getElements().iterator();
						while (i.hasNext()) {
							final Type current = i.next();
							TypeReferencer.this.handleType(current, clazz);
						}
					}
					
					@Override
					public void handle(final ProductType product) {
						TypeReferencer.this.getInputModel().getAnything().add(product);
						final Iterator<ProductElementType> i = product.getElements().iterator();
						while (i.hasNext()) {
							final ProductElementType current = i.next();
							TypeReferencer.this.handleType(current.getType(), clazz);
						}
					}
					
					@Override
					public void handle(final MapType map) {
						TypeReferencer.this.getInputModel().getAnything().add(map);
						TypeReferencer.this.handleType(map.getKey(), clazz);
						TypeReferencer.this.handleType(map.getOf(), clazz);
					}
					
					@Override
					public void handle(final ListType list) {
						TypeReferencer.this.getInputModel().getAnything().add(list);
						TypeReferencer.this.handleType(list.getOf(), clazz);
					}
					
					@Override
					public void handle(final ThrownType thrownType) {
						TypeReferencer.this.handleType(thrownType.getUnderlyingType(), clazz);
					}
				});
			}
			
			@Override
			public void handle(final AtomicType s) {
				throw new ASTException("Atomic Type is here not allowed!");
			}
		});
	}
	
	/**
	 * This method matches all applied TypeProxys in <code>appliedTypeProxys</code> <br/>
	 * against <code>definedClasses</code>.<br/>
	 * An Exception is saved in <code>exceptions</code> each time a matching fails.
	 */
	private void matchTypeProxysToClassTypes() {
		final Iterator<Tuple<TypeProxy, ClassType>> appliedTypeProxysIterator = this.getAppliedTypeProxys().iterator();
		while (appliedTypeProxysIterator.hasNext()) {
			final Tuple<TypeProxy, ClassType> current = appliedTypeProxysIterator.next();
			final TypeProxy currentReference = current.getFirst();
			final ClassType clazz = current.getSecond();
			
			currentReference.getState().accept(new TypeProxyStateVisitor() {
				
				@Override
				public void handle(final ByReferenceState byType) {
					// Nothing
				}
				
				@Override
				public void handle(final InvalidState invalid) {
					TypeReferencer.this.getExceptions().add(WrongStateException.create(invalid.toString()));
				}
				
				@Override
				public void handle(final ByNameState byName) {
					final Name appliedName = byName.getName();
					final Name classHierarchie =
							((QualifiedName) clazz.getName()).getNameWithoutLast(new ArrayList<IdentifierToken>());
					try {
						// try to find classtype with the applied name
						
						final Name newAppliedName =
								appliedName.visit(new NameVisitorReturnNameException<InvalidTypeReferenceException>() {
									@Override
									public Name handleUnqualifiedName(final UnqualifiedName unqName)
											throws InvalidTypeReferenceException {
										return TypeReferencer.this.findClassNameInGroupHierarchie(
												classHierarchie,
												unqName);
									}
									
									@Override
									public Name handleQualifiedName(final QualifiedName qName)
											throws InvalidTypeReferenceException {
										try {
											final Name newName = classHierarchie.concat(appliedName);
											newName.setFirstToken(classHierarchie.getFirstToken());
											newName.setLastToken(appliedName.getLastToken());
											final ClassType clazz2 = TypeReferencer.this.matchNameToClass(newName);
											return clazz2.getName();
										} catch (final InvalidTypeReferenceException e) {
											return TypeReferencer.this.findGroupNameInHierarchie(
													classHierarchie,
													appliedName);
										}
									}
								});
						TypeReferencer.this.matchAndChangeState(newAppliedName, currentReference);
					} catch (final InvalidTypeReferenceException e) {
						// classtype with the name was not found: So it reference a basetype or it is a wrong reference
						
						if (TypeReferencer.this.isString(appliedName)) {
							final ByReferenceState byType =
									ByReferenceState.create(
											TypeReferencer.this.getInputModel().getString(),
											appliedName);
							currentReference.changeState(byType);
						} else if (TypeReferencer.this.isInteger(appliedName)) {
							final ByReferenceState byType =
									ByReferenceState.create(
											TypeReferencer.this.getInputModel().getInteger(),
											appliedName);
							currentReference.changeState(byType);
						} else if (TypeReferencer.this.isAnything(appliedName)) {
							final ByReferenceState byType =
									ByReferenceState.create(
											TypeReferencer.this.getInputModel().getAnything(),
											appliedName);
							currentReference.changeState(byType);
						} else {
							TypeReferencer.this.processMatchingError(e, currentReference);
						}
					}
				}
			});
		}
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		this.matchTypeProxysToClassTypes();
		if (!this.getExceptions().isEmpty()) {
			throw ReferencingNotSuccessfulException.create();
		}
		
		// Fill All defined classes in List of Types for Anything. Their already
		// every Sum,Product,List and Map in it.
		for (final ClassType classType : this.getDefinedClasses()) {
			final ByReferenceState state = ByReferenceState.create(classType, classType.getTypeName());
			final TypeProxy typeProxy = TypeProxy.create(DummyToken.getInstance(), state, DummyToken.getInstance());
			this.inputModel.getAnything().add(typeProxy);
		}
		// Fill Basetypes in List of Types for Anything
		ByReferenceState state =
				ByReferenceState.create(this.inputModel.getString(), this.inputModel.getString().getTypeName());
		TypeProxy typeProxy = TypeProxy.create(DummyToken.getInstance(), state, DummyToken.getInstance());
		this.inputModel.getAnything().add(typeProxy);
		state = ByReferenceState.create(this.inputModel.getInteger(), this.inputModel.getInteger().getTypeName());
		typeProxy = TypeProxy.create(DummyToken.getInstance(), state, DummyToken.getInstance());
		this.inputModel.getAnything().add(typeProxy);
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do here
	}
	
	// ****** Helper-Methods ******//
	
	/**
	 * This method returns a {@link de.fhdw.wtf.common.ast.Name}, if a class is found matching to the applied name
	 * <code>className</code>.<br>
	 * Starting point for the recursive search is <code>groupPart</code> and if there is no match an Exception will be
	 * thrown.
	 * 
	 * @param groupPart
	 *            : Name
	 * @param className
	 *            : UnqualifiedName
	 * @return : Name.
	 * @throws InvalidTypeReferenceException
	 *             possible Exception
	 */
	protected Name findClassNameInGroupHierarchie(final Name groupPart, final UnqualifiedName className)
			throws InvalidTypeReferenceException {
		return groupPart.visit(new NameVisitorReturnNameException<InvalidTypeReferenceException>() {
			@Override
			public Name handleUnqualifiedName(final UnqualifiedName unqName) throws InvalidTypeReferenceException {
				final Name newName = unqName.concat(className);
				newName.setFirstToken(className.getLastToken());
				newName.setLastToken(className.getLastToken());
				TypeReferencer.this.matchNameToClass(newName);
				return newName;
			}
			
			@Override
			public Name handleQualifiedName(final QualifiedName qName) throws InvalidTypeReferenceException {
				final Name newName = qName.concat(className);
				newName.setFirstToken(qName.getFirstToken());
				newName.setLastToken(className.getLastToken());
				try {
					TypeReferencer.this.matchNameToClass(newName);
					return newName;
				} catch (final InvalidTypeReferenceException e) {
					final Name newGroupPart = qName.getNameWithoutLast(new ArrayList<IdentifierToken>());
					return TypeReferencer.this.findClassNameInGroupHierarchie(newGroupPart, className);
				}
			}
		});
	}
	
	/**
	 * This method returns a {@link de.fhdw.wtf.common.ast.Name}, if a class is found matching to the applied name
	 * <code>appliedName</code>.<br/>
	 * Starting point for the recursive search is <code>groupPart</code> and if there is no match an Exception will be
	 * thrown.
	 * 
	 * @param groupPart
	 *            : Name.
	 * @param appliedName
	 *            : Name.
	 * @return : Name.
	 * @throws InvalidTypeReferenceException
	 *             possible Exception
	 */
	protected Name findGroupNameInHierarchie(final Name groupPart, final Name appliedName)
			throws InvalidTypeReferenceException {
		if (groupPart.equalsPartial(appliedName)) {
			return groupPart.visit(new NameVisitorReturnName() {
				@Override
				public Name handleUnqualifiedName(final UnqualifiedName unqName) {
					return appliedName;
				}
				
				@Override
				public Name handleQualifiedName(final QualifiedName qName) {
					final Name part = qName.getNameWithoutLast(new ArrayList<IdentifierToken>());
					final Name resultName = part.concat(appliedName);
					resultName.setFirstToken(part.getFirstToken());
					resultName.setLastToken(appliedName.getLastToken());
					return resultName;
				}
			});
		} else {
			return groupPart.visit(new NameVisitorReturnNameException<InvalidTypeReferenceException>() {
				@Override
				public Name handleUnqualifiedName(final UnqualifiedName unqName) throws InvalidTypeReferenceException {
					throw InvalidTypeReferenceException.create(appliedName, unqName.getIdentifierToken());
				}
				
				@Override
				public Name handleQualifiedName(final QualifiedName qName) throws InvalidTypeReferenceException {
					return TypeReferencer.this.findGroupNameInHierarchie(
							qName.getNameWithoutLast(new ArrayList<IdentifierToken>()),
							appliedName);
				}
			});
		}
	}
	
	/**
	 * This method tries to match the {@link de.fhdw.wtf.common.ast.Name} <code>toMatch</code> to a defined
	 * {@link ClassType} and changes the state of the {@link TypeProxy} <code>reference</code> to
	 * {@link ByReferenceState} at success. An Exception is thrown if there is no match.
	 * 
	 * @param toMatch
	 *            : The Name to be matched.
	 * @param typeProxy
	 *            : The current TypeProxy.
	 * @throws InvalidTypeReferenceException
	 *             If there is no match.
	 */
	protected void matchAndChangeState(final Name toMatch, final TypeProxy typeProxy)
			throws InvalidTypeReferenceException {
		final ClassType target = this.matchNameToClass(toMatch);
		final ByReferenceState byType = ByReferenceState.create(target, toMatch);
		typeProxy.changeState(byType);
	}
	
	/**
	 * This method takes a {@link Name} and tries to match it to a {@link Name} of a defined {@link ClassType}.
	 * 
	 * @param appliedName
	 *            : The given {@link Name}.
	 * @return : The matching {@link ClassType}.
	 * @throws InvalidTypeReferenceException
	 *             : If no matching {@link ClassType} is found.
	 */
	protected ClassType matchNameToClass(final Name appliedName) throws InvalidTypeReferenceException {
		boolean found = false;
		ClassType target = null;
		final Iterator<ClassType> classes = this.getDefinedClasses().iterator();
		while (classes.hasNext()) {
			final ClassType current = classes.next();
			if (appliedName.equals(current.getName())) {
				found = true;
				target = current;
				break;
			}
		}
		if (!found) {
			throw InvalidTypeReferenceException.create(appliedName, appliedName.getLastAddedName().getFirstToken());
		}
		return target;
	}
	
	/**
	 * This method adds a given {@link InvalidTypeReferenceException} to the list of exceptions of this referencer.
	 * Furthermore it changes the state of the given {@link TypeProxy} to {@link InvalidState}.
	 * 
	 * @param e
	 *            : The given {@link InvalidTypeReferenceException}.
	 * @param typeProxy
	 *            : The given {@link TypeProxy}.
	 */
	protected void processMatchingError(final InvalidTypeReferenceException e, final TypeProxy typeProxy) {
		this.getExceptions().add(e);
		final InvalidState invalidState = InvalidState.create(e.getName());
		typeProxy.changeState(invalidState);
	}
	
	/**
	 * Returns true if the given {@link Name} <code>name</code> matches to a string representation of String specified
	 * in {@link de.fhdw.wtf.common.constants.referencer.BaseTypeConstants}.
	 * 
	 * @param name
	 *            : The given {@link Name}.
	 * @return : boolean.
	 */
	protected boolean isString(final Name name) {
		return name.toString().equals(BaseTypeConstants.STRING_CONSTANT);
	}
	
	/**
	 * Returns true if the given {@link Name} <code>name</code> matches to a string representation of Integer specified
	 * in {@link de.fhdw.wtf.common.constants.referencer.BaseTypeConstants}.
	 * 
	 * @param name
	 *            : The given {@link Name}.
	 * @return : boolean.
	 */
	protected boolean isInteger(final Name name) {
		return name.toString().equals(BaseTypeConstants.INTEGER_CONSTANT);
	}
	
	/**
	 * Returns true if the given {@link Name} matches the string representation of Anything.
	 * 
	 * @param name
	 *            : The given {@link Name} to check
	 * @return : boolean
	 */
	protected boolean isAnything(final Name name) {
		return name.toString().equals(BaseTypeConstants.ANYTHING_CONSTANT);
	}
	
	// ****** PROJECTIONS ******//
	
	/**
	 * Returned the model.
	 * 
	 * @return model
	 */
	protected Model getInputModel() {
		return this.inputModel;
	}
	
	/**
	 * Returned the Collection of Exceptions.
	 * 
	 * @return exceptions
	 */
	public Collection<TaskException> getExceptions() {
		return this.exceptions;
	}
	
	/**
	 * Returned the found Type Proxys.
	 * 
	 * @return typeProxys
	 */
	protected List<Tuple<TypeProxy, ClassType>> getAppliedTypeProxys() {
		return this.appliedTypeProxys;
	}
	
	/**
	 * Returned the found classes.
	 * 
	 * @return classes
	 */
	protected List<ClassType> getDefinedClasses() {
		return this.definedClasses;
	}
	
}
